
package snap.adapter;

import snap.*;
import snap.database.*;

import java.util.*;
import java.sql.*;
import java.net.URL;
import javax.swing.*;

/** Adapter between Snap's general interface to the interface specific
  * for Java visualizations.  This adapter makes the event multi-threaded by
  * implementing a thread for event visualization.  All events are queued and
  * fired in FIFO manner.  The adapter adds a "Load" action as a supported
  * event.  Any "Load" SnapEvents are converted into a call to the Snapable
  * interface's load() method.
  *
  * @author Nathan Conklin - nathan@conklinfamily.net
  * @author Varun Saini - vsaini@vt.edu 
  * @author Chris North - north@cs.vt.edu
  * @author Umer Farooq - ufarooq@vt.edu
  */
public class JavaTechnologyAdapter extends TechnologyAdapter
implements SnapEventListener, Runnable
{

private Thread t;
private Snap snap;
private Snapable visualization;
private SnapEventListener sel;

/* Set when firing an event at a visualization.  When set, the adapter
ignores events coming back from the visualization.  Allows the adapter to
ignore talkback. */
private boolean ignoreVisualizationEvents;

private final int HANDLE_LOAD = 0;
private final int HANDLE_PERFORMSNAPEVENT = 1;
private final int HANDLE_SNAPEVENTOCCURED = 2;

private final String LOAD_ACTION = "Load";

private Vector loadQueue = new Vector();
private Vector snapEventOccuredQueue = new Vector();
private Vector performSnapEventQueue = new Vector();
private Vector eventSelectionQueue = new Vector();

// In JDBC, each thread should have its own connection
private Connection connection = null;

/** Constructs the JavaTechnologyAdapter with a reference to the
Java-based visualization that it will talk to. */
public JavaTechnologyAdapter(Snap snap, Snapable visualization)
{
	t = new Thread(this);
    this.visualization = visualization;
	this.snap = snap;
	visualization.addSnapEventListener(this);
	t.start();
}

/** Returns the component's Icon or tries to find a default Icon */
public Icon getIcon()
{
	Icon icon = visualization.getIcon();
	if (icon == null)
	{
		try
		{
			// Must change
			String base = snap.getCodeBase().toString();
			String iconURL = base + "snap/hotjava.icon.gif";
			icon = 	new ImageIcon(new URL(iconURL));
		}
		catch(Exception e)
		{
			System.err.println("JavaTechnologyAdapter " + 
				"unable to create an icon: " + e.toString());
		}
	}
	return icon;
}

/** Enqueues the Query to be loaded by the visualization */
public void load(Query query)
{
	loadQueue.add(query);
	interruptThread(HANDLE_LOAD);
}

/** Dequeues the Query then calls the executeLoad */
private void handleLoad()
{
	if(loadQueue.size() > 0)
	{
		Query query = (Query) loadQueue.remove(0);
		executeLoad(query);
	}
}


/** Enqueues the SnapEvent to be performed by the visualization */
public void performSnapEvent(SnapEvent event)
{
	performSnapEventQueue.add(event);
	interruptThread(HANDLE_PERFORMSNAPEVENT);
}

/** Dequeues the SnapEvent to be performed by the visualization */
private void handlePerformSnapEvent()
{
	if(performSnapEventQueue.size() > 0)
	{
		SnapEvent event = (SnapEvent) performSnapEventQueue.remove(0);
		executeHandlePerformSnapEvent(event);
	}
}

/** Enqueues the SnapEvent to be performed by Snap */
public void snapEventOccured(SnapEvent e)
{
	if(!ignoreVisualizationEvents) // Used to ignore component talkback
	{
		SnapEvent event =
			new SnapEvent(this , e.getEventType(), e.getKeys());
		snapEventOccuredQueue.add(event);
		interruptThread(HANDLE_SNAPEVENTOCCURED);
	}
}

/** Dequeues the SnapEvent to fire back to Snap */
private void handleSnapEventOccured()
{
	if(snapEventOccuredQueue.size() > 0)
	{
		SnapEvent event = (SnapEvent) snapEventOccuredQueue.remove(0);
		fireSnapEvent(event);
	}
}


/** Returns the list of Actions supported by the visualization */
public Enumeration getSupportedActions()
{
	Enumeration e = visualization.getSupportedActions();
	Vector v = new Vector();
	v.add(LOAD_ACTION);
	while (e.hasMoreElements())
	{
		v.add((String) e.nextElement());
	}
    return v.elements();
}

/** Calls the load() method on the visualization */
private void executeLoad(Query query)
{
	try
	{
		ignoreVisualizationEvents = true;

		String table = query.getTable();
		DatabaseManager dbm = snap.getDatabaseManager();
		DatabaseSchema schema = dbm.getDatabaseSchema();
		if (connection == null)
			connection = dbm.getNewConnection();
		ResultSet data = dbm.getResultSet(connection, query);
		visualization.load(data, schema.getPrimaryKey(table));

		ignoreVisualizationEvents = false;
	}
	catch(Exception e)
	{
		snap.getFeedbackPanel().setText(e.toString());
		e.printStackTrace();
	}
}

/** Transforms a "Load" SnapEvent into a load() method call */
private void executeSnapEventLoad(SnapEvent event)
{
	String table = getNode().getQuery().getTable();
	DatabaseManager dbm = snap.getDatabaseManager();
	DatabaseSchema schema = dbm.getDatabaseSchema();
	Query filteredQuery = new Query(getNode().getQuery());

	if (event.getKeys().size() > 0)
	{
		String whereClause = schema.getPrimaryKey(table) + " IN (";
		Enumeration e = event.getKeys().elements();
		while (e.hasMoreElements())
		{
			Object key = e.nextElement();
			if (key.getClass().equals(	// Checks if the primary key is a String
				(new String()).getClass()))
				key = "'" + key.toString() + "'";

			whereClause += key;
			if (e.hasMoreElements())
				whereClause += ", ";
		}
		whereClause += " )";
		filteredQuery.addToWhereClause(whereClause);
	}
	else
		filteredQuery.addToWhereClause(
			schema.getPrimaryKey(table) + " = NULL");
	executeLoad(filteredQuery);
}

/** Calls the performSnapEvent method on the visualization */
private void executeHandlePerformSnapEvent(SnapEvent event)
{
	if (LOAD_ACTION.equals(event.getEventType()))  // Load Event
	{
		executeSnapEventLoad(event);
	}
	else // Regular SnapEvent
	{
		try
		{
			ignoreVisualizationEvents = true;
			visualization.performSnapEvent(event);
			ignoreVisualizationEvents = false;
		}
		catch(Exception e)
		{
			snap.getFeedbackPanel().setText(
				"Error talking to " + getNode().getName() +
				"\n" + e.toString());
			e.printStackTrace();
		}
	}
}


/** Interrupts the adapter thread */
private void interruptThread(int reasonForInterruption)
{
	try
    {
		eventSelectionQueue.add(new Integer(reasonForInterruption));
        t.interrupt();
    }
    catch(SecurityException e)
    {
		System.err.println(e.toString());
		e.printStackTrace();
    }
}

/** Handles an event when the thread is interrupted */
private void handleAnyEvents()
{
	while(eventSelectionQueue.size() > 0)
	{
		int reasonForInterruption;
		Integer i = (Integer) eventSelectionQueue.remove(0);
		reasonForInterruption = i.intValue();
		switch(reasonForInterruption)
		{
			case HANDLE_LOAD:
				handleLoad();
				break;
			case HANDLE_PERFORMSNAPEVENT:
				handlePerformSnapEvent();
				break;
			case HANDLE_SNAPEVENTOCCURED:
				handleSnapEventOccured();
				break;
			default:
				System.err.println("Thread interrupted");
				break;
		}
	}
}

/** Run's the JavaTechnologyAdapter in its own thread. */
public void run()
{
    while(true)
    {
        try
		{
			t.sleep(30000);
		}
		catch(InterruptedException e)
		{
			handleAnyEvents();
			// Manage the queue
		}
    }
}

}
