package snap.coordination;

import java.sql.*;
import snap.*;
import snap.database.*;
import java.util.*;

/**
  * Responsible for converting the keys between events.
  * It takes the originating snap event and after the 
  * key conversion, returns a new snap event.
  *
  * @author Varun Saini - vsaini@vt.edu 
  * @author Chris North - north@cs.vt.edu
  * @author Nathan Conklin - nathan@conklinfamily.net
  * @author Umer Farooq - ufarooq@vt.edu
  */
  
public class KeyConverter implements ConnectionChangeEventListener
{

// Access the database and retrieves the schema
private DatabaseManager dbm;
private DatabaseSchema schema;

/**
 * Constructor.
 */
public KeyConverter(DatabaseManager manager)
{
	dbm = manager;
	manager.addConnectionChangeEventListener(this);
}

/**
 * This is the event which gets the DatabaseSchema.
 */
public void connectionChanged(ConnectionChangeEvent e)
{
	schema = ((DatabaseManager)e.getSource()).getDatabaseSchema();
}

/**
 * Method that converts the keys and returns a new
 * snap event.
 */
public SnapEvent convertKeys(Query srcQuery, Query destQuery, SnapEvent se)
{
	// Check if we have handle to database schema
	if(schema == null || srcQuery.isEmpty() || destQuery.isEmpty())
		return se;

	// Key values in source table to be converted
	Vector values = se.getKeys();

	// If no keys are selected in source, then return
	// a snap event in which no keys are selected for
	// destination
	if (values == null || values.size() == 0)
	{
		SnapEvent toReturn= new SnapEvent(this, null, new Vector());
		return toReturn;
	}

	// Get source table name
	String tableSrc = srcQuery.getTable();
	// Get destination table name
	String tableDest = destQuery.getTable();

	String keySrc, keyDest;

	// Handle the key conversion based on relationship
	switch (schema.getRelation(tableSrc, tableDest))
	{
		case DatabaseSchema.ONE_TO_ONE:
			return se;

		case DatabaseSchema.MANY_TO_ONE:
			keySrc = schema.getPrimaryKey(tableSrc);
			keyDest = schema.getPrimaryKey(tableDest);
			Query srcCopy = new Query(srcQuery);
			return fireQuery(srcCopy, values, keySrc, keyDest);

		case DatabaseSchema.ONE_TO_MANY:
			keySrc = schema.getPrimaryKey(tableSrc);
			/*String tempTable = "";
			while (!tempTable.equals(tableDest))
			{
				tempTable = 
schema.getRelatedTableOnTheWayToDestTable(tableDest);
			}*/
			keyDest = schema.getPrimaryKey(tableDest);
			Query destCopy = new Query(destQuery);
			return fireQuery(destCopy, values, keySrc, keyDest);
			
		case DatabaseSchema.NO_RELATION:
			return se;
	}

	System.err.println("Problem with Database Schema");
	return se;
}



/**
 * Creates the query for 1-M and M-1 key conversion.
 */

private SnapEvent fireQuery(Query query, Vector values,
	String keySrc, String keyDest)
{
	// Where clause of the final query to be executed
	// "keySrc IN ( value1, value2, value3 )"
	String whereClause = keySrc + " IN ( ";
	
	Enumeration e = values.elements();
	while (e.hasMoreElements())
	{	
		Object Obj = e.nextElement();
		System.out.println("NEW KEY CONVERTER " + Obj.getClass().getName());
		if ( Obj.getClass().getName().equals("java.util.Date")  
			|| Obj.getClass().getName().equals("java.lang.String") ||
			 Obj.getClass().getName().equals("java.sql.Time")) {  
			
			whereClause += "'" + Obj + "'"; 
		}
		else {
			
				whereClause += Obj ; 
		}
		
		if (e.hasMoreElements())
			whereClause += ", ";
	}
	whereClause += " )";

	// Creates the copy of src query because we don't want
	// to modify the src query (adding a filter)
	Query toFire = new Query(query);

	// Add where clause to the query to be fired
	toFire.addToWhereClause(whereClause);

	System.out.println(toFire.toString());

	// Fire the query
	try
	{
		Connection conn = dbm.getConnection();
		Statement stmt = conn.createStatement();
		ResultSet rs = stmt.executeQuery(toFire.toString());
		values = new Vector();

		// Getting the destination 	
		// primary key values from the resultset
		while (rs.next())
		{
			values.add(rs.getObject(keyDest));
		}
	}
	catch (Exception exception)
	{
		System.out.println("Exception in convertKeys method");
		exception.printStackTrace();
	}
	
	// Snap event to be returned
	SnapEvent toReturn= new SnapEvent(this, null, values);
	return toReturn;
}

}
