
package snap.database;

import snap.coordination.*;
import snap.*;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.awt.event.*;

import java.util.*;
import java.sql.*;
import java.net.InetAddress;

/** Manages the database connections
  *
  * @author Nathan Conklin - nathan@conklinfamily.net
  * @author Chris North - north@cs.vt.edu
  */
public class DatabaseManager extends JPanel
implements ActionListener
{

protected Snap snap;

protected Connection dbServerConnection;

protected JButton connectButton = new JButton("Connect");
protected JComboBox server;
protected JPanel settings = new JPanel();
protected CardLayout cardLayout = new CardLayout();

protected JTextField database = new JTextField();
protected JTextField user = new JTextField();
protected JPasswordField password = new JPasswordField();
protected JTextField query = new JTextField();

protected String noSystemDatabasesMessage = "No Shared Databases";
protected JComboBox databaseChoice;
protected JLabel databaseDescription = new JLabel();
protected Vector systemDatabases = new Vector();

// Default database to connect to on the localhost
//protected String defaultDSN = "Data.dsn";
protected String defaultDSN = "Molecules.dsn";
protected String defaultUser = "";
protected String defaultPassword = "";

/* Information used to connect to the server database.  This is the
database that contains information about system databases available
for visualization */
private String snapDatabase = "Snap.dsn";
private String snapDatabaseUser = "";
private String snapDatabasePassword = "";


// java.sql components used in opening the datasource
protected Connection dataConnection = null;
private DatabaseSchema dataSchema = null;

// Container to keep track of the registered listeners
private Vector listeners = new Vector();

/** Removes the listener for ConnectionChangeEvents */
public void removeConnectionChangeEventListener(
	ConnectionChangeEventListener ccel)
{
    listeners.remove(ccel);
}

/** Installs the listener for ConnectionChangeEvents */
public void addConnectionChangeEventListener(ConnectionChangeEventListener ccel)
{
    listeners.add(ccel);
}

/** Fires a ConnectionChangeEvent */
protected void fireConnectionChangeEvent(ConnectionChangeEvent cce)
{
	Enumeration e = listeners.elements();
	while(e.hasMoreElements())
	{
		ConnectionChangeEventListener listener;
		listener = (ConnectionChangeEventListener) e.nextElement();
		listener.connectionChanged(cce);
	}
}

/** Constructor for the DatabaseManager.
  * @param applet is a reference to the Snap object **/
public DatabaseManager(Snap snap)
{
    this.snap = snap;
    jbInit();
}

public void actionPerformed(ActionEvent e)
{
    if(e.getSource() == connectButton)
        connectToDataSource();
    else if(e.getSource() == server)
        showDatabaseConfiguration();
    else if(e.getSource() == databaseChoice)
        setDatabaseDescription();
}


/** This method will return a database connection */
public Connection getConnection(String server, String database,
                                String user, String password)
{
    String url;
    Driver driver = null;
    Connection connection = null;

    // JDBC Drivers that I have tested but am not using
    // url = "jdbc:rmi://" + server + "/jdbc:odbc:" + database;
    // url = "jdbc:freetds:sqlserver://" + server + "/" + database;
    // driverClass = "RmiJdbc.RJDriver";
    // driverClass = "com.internetcds.jdbc.tds.Driver";

    if(server.equals("localhost"))
        url = "jdbc:odbc:" + database;
    else
        url = "jdbc:JDataConnect://" + server + "/" + database;

    try // Register JDBC Driver
    {
		String driverClass;
		if(server.equals("localhost"))
			driverClass = "sun.jdbc.odbc.JdbcOdbcDriver";
		else
			driverClass = "JData2_0.sql.$Driver";
		driver = (Driver) Class.forName(driverClass).newInstance();
    }
    catch (Exception e)
    {
        String err = "Cannot load the driver, reason:\n" + e.toString(); 
		if(snap.getFeedbackPanel() == null)
			System.err.println(err);
		else
			snap.getFeedbackPanel().setText(err);
		return null;
    }

    try // Establish a connection
    {
		if (user.length() > 0)
			connection = DriverManager.getConnection(url, user, password);
		else
			connection = DriverManager.getConnection(url);
		return connection;
    }
    catch(SQLException e)
    {
		String err = "Failed to connect to " + server + "/" +
			database + "\n" + "SQLException: " + e.toString();
		if(snap.getFeedbackPanel() == null)
			System.err.println(err);
		else
			snap.getFeedbackPanel().setText(err);
		return null;
    }
}

/** Returns a new connection to the user's database.  Returns null if Snap
  * was unable to connect.  This method is needed to allow each thread to
  * have its own connection to the database. */
public Connection getNewConnection()
{
    return getConnection(getServer(), getDSN(), getUser(), getPassword());
}

/** Returns the connection to the user's database.  null if Snap
was unable to connect */
public Connection getConnection()
{
    return dataConnection;
}

/** Returns the DatabaseSchema for the data connection.  null if Snap
was unable to connect */
public DatabaseSchema getDatabaseSchema()
{
    return dataSchema;
}

/** This method will return the ResultSet for a given SQL statement.
    @return ResultSet for the given connection and SQL statement.
    null if there was a problem retrieving the ResultSet. */
public ResultSet getResultSet(Connection connection, Query sql)
{
    Statement statement;
    ResultSet rs;

    try
    {
        statement = connection.createStatement(
			ResultSet.TYPE_SCROLL_INSENSITIVE,
			ResultSet.CONCUR_READ_ONLY);
		//connection.createStatement();
        rs = statement.executeQuery(sql.toString());
    }
    catch(SQLException e)
    {
		String err = "Error retrieving ResultSet: " + e.toString();
		if(snap.getFeedbackPanel() == null)
			System.err.println(err);
		else
			snap.getFeedbackPanel().setText(err);
        rs = null;
    }
    return rs;
}

/** Returns the server that has been set by the user. **/
public String getServer()
{
    return (String) server.getSelectedItem();
}

/** Returns the database name that has been set by the user. **/
public String getDSN()
{
    return database.getText();
}

/** Returns the user name that has been set by the user. **/
public String getUser()
{
    return user.getText();
}

/** Returns the password that has been set by the user. **/
public String getPassword()
{
    char[] passArray = password.getPassword();
    String passString = new String(passArray);
    return passString;
}



private void jbInit()
{
    String[] choices;
    if(snap.getHost() == null || snap.getHost().equals(""))
    {
       choices = new String[2];
       choices[1] = "localhost";
       choices[0] = "infovis.cs.vt.edu";
       this.dbServerConnection = getConnection("infovis.cs.vt.edu", "Molecules.dsn", snapDatabaseUser,  snapDatabasePassword);
	 // this.dbServerConnection = null;
    }
    else
    {
		choices = new String[2];
		this.dbServerConnection = getConnection("infovis.cs.vt.edu",
			"Molecules.dsn", snapDatabaseUser, snapDatabasePassword);
		choices[1] = snap.getHost();
		choices[1] = "localhost";
	        choices[0] = "infovis.cs.vt.edu";
    }
    server = new JComboBox(choices);
    server.setSelectedIndex(0);

    database.setText(defaultDSN);
    user.setText(defaultUser);
    password.setText(defaultPassword);
    //query.setText(defaultQuery);

    JPanel serverChoicePanel = new JPanel();
    serverChoicePanel.setLayout(new GridLayout(1,2));
    serverChoicePanel.add(new JLabel("Database Server:"));
	serverChoicePanel.setBorder(BorderFactory.createEmptyBorder(0,5,0,0));
    serverChoicePanel.add(server);

    JPanel userConfig = new JPanel();
    userConfig.setLayout(new GridLayout(3,2));
    userConfig.add(new JLabel("Database Name:"));
    userConfig.add(database);
    userConfig.add(new JLabel("User:"));
    userConfig.add(user);
    userConfig.add(new JLabel("Password:"));
    userConfig.add(password);
	userConfig.setBorder(BorderFactory.createEmptyBorder(0,5,0,0));

    JPanel serverDatabasePanel = new JPanel();
    setupDatabaseChoice();
    serverDatabasePanel.setLayout(new GridLayout(3,2));
	serverDatabasePanel.add(new JLabel("Database Name:"));
    serverDatabasePanel.add(databaseChoice);

	//JPanel serverDatabasePanel = new JPanel(new BorderLayout());
    setDatabaseDescription();
	//serverDatabasePanel.add(serverDatabaseNamePanel, BorderLayout.NORTH);
	serverDatabasePanel.add(new JLabel("Database Description:"));
    serverDatabasePanel.add(databaseDescription);
	databaseDescription.setBorder(BorderFactory.createEmptyBorder(0,3,0,0));
	serverDatabasePanel.setBorder(BorderFactory.createEmptyBorder(0,5,0,0));

    settings.setLayout(cardLayout);
    settings.add("System Database", serverDatabasePanel);
    settings.add("User Database", userConfig);
    JPanel settingsPanel = new JPanel();
    settingsPanel.setLayout(new BorderLayout());
    settingsPanel.add(settings, BorderLayout.NORTH);

    JPanel totalConfig = new JPanel();
    totalConfig.setLayout(new BorderLayout());
    totalConfig.add(serverChoicePanel, BorderLayout.NORTH);
    totalConfig.add(settingsPanel, BorderLayout.CENTER);
	//totalConfig.setMaximumSize(new Dimension(0, 350));

	JPanel connectPanel = new JPanel();
//    connectPanel.setLayout(new BorderLayout());
    connectPanel.add(connectButton);

	JPanel dbManagerPanel = new JPanel();
    dbManagerPanel.setLayout(new BorderLayout());
    dbManagerPanel.add(totalConfig, BorderLayout.NORTH);
    dbManagerPanel.add(connectPanel, BorderLayout.CENTER);
    JScrollPane scrollPane = new JScrollPane(dbManagerPanel);

	this.setLayout(new BorderLayout());
	this.add(scrollPane, BorderLayout.CENTER);

    server.addActionListener(this);
    connectButton.addActionListener(this);
    databaseChoice.addActionListener(this);
	showDatabaseConfiguration();
}


/** Sets the database description for the chosen database */
private void setDatabaseDescription()
{
	if (systemDatabases.size() > 0)
	{
		int selected = databaseChoice.getSelectedIndex();
		String[] config = (String[]) systemDatabases.elementAt(selected);
		if(config != null)
		{
			databaseDescription.setText(config[1]);
			database.setText(config[3]);
			user.setText(config[4]);
			password.setText(config[5]);
		}
	}
}

private void showDatabaseConfiguration()
{
    String serverText = (String) server.getSelectedItem();
    if(serverText.equals(snap.getHost()))
    {
		setDatabaseDescription();
        cardLayout.show(settings, "System Database");
    }
    else
    {
        cardLayout.show(settings, "User Database");
    }
}

/** Adds the database settings into the drop down menu */
private void setupDatabaseChoice()
{
	if (dbServerConnection != null)
	{
		try
		{
			Statement s = dbServerConnection.createStatement();
			String query = "SELECT * FROM SystemDatabases " +
			"ORDER BY DatabaseName";
			ResultSet rs = s.executeQuery(query);
			while(rs.next())
			{
				String[] databaseConfig = new String[6];
				databaseConfig[0] = rs.getString("DatabaseName");
				databaseConfig[1] = rs.getString("DatabaseDescription");
				databaseConfig[2] = rs.getString("Server");
				databaseConfig[3] = rs.getString("Name");
				databaseConfig[4] = rs.getString("Username");
				databaseConfig[5] = rs.getString("Password");
				systemDatabases.add(databaseConfig);
			}
		}
		catch(Exception e)
		{
			String err = e.toString() + "\nUnable to find system databases.";
			if (snap.getFeedbackPanel() == null)
				System.err.println(err);
			else
				snap.getFeedbackPanel().setText(err);
		}

		String[] choices;
		if(systemDatabases.size() > 0)
		{
			choices = new String[systemDatabases.size()];

			Enumeration e = systemDatabases.elements();
			int i = 0;
			while(e.hasMoreElements())
			{
				String[] temp = (String[]) e.nextElement();
				choices[i] = temp[0];
				i++;
			}
		}
		else
		{
			choices = new String[1];
			choices[0] = noSystemDatabasesMessage;
		}
		databaseChoice = new JComboBox(choices);
		databaseChoice.setLightWeightPopupEnabled(false);
		databaseChoice.setSelectedIndex(0);
	}
	else
		databaseChoice = new JComboBox();
}


/** Once all of the connection properties have been set by
  * the UI, connectToDataSource() will establish the connection
  * to the database **/
private void connectToDataSource()
{
    if(dataConnection != null)
    {
		try
		{
			dataSchema = null;
			dataConnection.close();
			dataConnection = null;
			fireConnectionChangeEvent(new ConnectionChangeEvent(this));
		}
		catch(Exception e)
		{
			snap.getFeedbackPanel().setText("Error closing the database: " +
				e.toString());
		}
		dataConnection = null;
    }

    dataConnection = getConnection(getServer(), getDSN(),
        getUser(), getPassword());

    if(dataConnection != null)
    {
		dataSchema = new DatabaseSchema(dataConnection);
		snap.getFeedbackPanel().setText("Connected to: " + getServer() +
			":" + getDSN());
    }
	fireConnectionChangeEvent(new ConnectionChangeEvent(this));
}




};

