package snap.database;

import snap.ui.MsgBox;

import java.sql.*;
import java.util.*;

/** This class retrieves schema information from the database.
  *
  * @author Varun Saini - vsaini@vt.edu
  * @author Chris North - north@cs.vt.edu
  * @author Nathan Conklin - nathan@conklinfamily.net
  */
public class DatabaseSchema
{

// Constants for relationships
public final static int NO_RELATION = 0;
public final static int ONE_TO_ONE = 1;
public final static int ONE_TO_MANY = 2;
public final static int MANY_TO_ONE = 3;

private Connection connection;
private DatabaseMetaData dmd;
private String catalog;					// Connection's Catalog
private String tableTypes[]={"TABLE"};	// Could be: {"TABLE", "VIEW"};
private ResultSet rs;

private Vector tables = new Vector();				// Table names
private Hashtable primaryKeys = new Hashtable();	// PK's for each table
private Hashtable columns = new Hashtable();		// Columns for each table

// Hashtable to maintain table relations
Hashtable oneToManyTableRelationship = new Hashtable();
Hashtable oneToOneTableRelationship = new Hashtable();

private int [][] parents; // An array containing the index of all parents 
private int [][] siblings; //An array containing the names of all siblings 
private int [][] children; //An array containing the names of all children 

private String[] newTables;  //Same as above with unconnected tables removed

private String[] unConnected; //List of unconnected tables

//private String[][] colnames; //Names of columns for each tables
private String[][] relations; //the connections matrix

private String[][] newRelations; //old relations=unRelated

private int numConnectedTables; // number of tables which are connected
private int root; //index of root table with max parents
private int[] numRows; //Number of rows for a table
private int[] numColumns;//Number of columns for a particular table

// Hack to catch errors like getCrossReference() not in Access
private boolean schemaCreated;

/**
 * This is the default constructor for the DatabaseSchema object
 * @param server The database server to connect to 
 * @param database The database name in the server 
 */
public DatabaseSchema(Connection connection)
{
	this.connection = connection;
	int[] referencesTo;
	int[] referencesFrom;

    try
    {
		// Get the MetaData
		dmd = connection.getMetaData();
		catalog = connection.getCatalog();

		setTables();
		setColumns();
		setPrimaryKeys();

		// Now we find cross references
			// System.out.print(
			//	"Retrieving schema information. This might take a while.");
		relations = new String[getTables().size()][getTables().size()];
		referencesTo = new int[getTables().size()];
		referencesFrom = new int[getTables().size()];
		for (int i=0; i < getTables().size(); i++)
		{
			referencesTo[i] = 0;
			referencesFrom[i] = 0;
		}
	
		int tableCount=0;
		while (tableCount < getTables().size())
		{
			int bb=0;
			while (bb < getTables().size())
			{
				String primaryTable = (String)tables.elementAt(tableCount);
				String foreignTable = (String)tables.elementAt(bb);

				rs = dmd.getCrossReference(
					catalog, null, primaryTable,
					catalog, null, foreignTable);
				while(rs.next())
				{
				// Let us assume that PKColName and FKColname are same. This
				// will simplify SQL Query generation
					relations[tableCount][bb] = rs.getString("PKCOLUMN_NAME");
					referencesTo[tableCount]++; 
				//	System.out.println(tableCount +" HAS A REF TO");
					referencesFrom[bb]++; 
				//	System.out.println(bb +" HAS A REF FROM");
				}
				
				bb++;
			}
			tableCount++;
		}	// while
 

		int countNum=0;
		Vector v3 = new Vector();
		boolean deadTable[]= new boolean [getTables().size()];
		for(int i=0; i < getTables().size(); i++)
		{
			if (!((referencesTo[i]==0) && (referencesFrom[i]==0)))
			{
				countNum++;
				v3.add((String)tables.elementAt(i));
				deadTable[i]=false;
			}
			else
			{	//System.out.println("DEAD :" +i);
				deadTable[i]=true;
			}	
		}// end for

		newTables = new String[countNum];
		numConnectedTables = countNum;
		//System.out.println("New Tables :" +numConnectedTables);
		v3.copyInto(newTables);

		newRelations= new String[numConnectedTables][numConnectedTables];
		tableCount = 0;
		int anew = 0;
		while (tableCount < getTables().size())
		{
			int bb=0;
			int bnew=0;
			if(!deadTable[tableCount])
			{
				while (bb<getTables().size())
				{
					
					if (!deadTable[bb])
					{
						newRelations[anew][bnew] = relations[tableCount][bb];
						bnew++;

					}
					bb++;
				}
				anew++;
			}	
			tableCount++;
		}

		// Relations is now DONE
		// Next we get the parents and children

		parents = new int[numConnectedTables][];
		// get one dimension of each of these arrays 
		// fixed now
		siblings= new int[numConnectedTables][];
		children= new int[numConnectedTables][];
		// System.out.println(getTables().size());
					
		int numParents[]= new int[numConnectedTables];
		// This array tells me the number of 
		// parents each table has.


		// This is where we find the siblings for each table
		// Right now, I just add each table as a sibling of itself
		for(int i=0; i < numConnectedTables; i++)
		{
			siblings[i]= new int[1];
			siblings[i][0]=i;
		}

		// Siblings is now DONE : It means I'm done

		// We simply traverse the relations array and build parents and siblings
		tableCount=0; // our counter for rows
		while (tableCount < numConnectedTables)
		{	
			int bb=0; //our counter for columns
			int numelts=0; //count the number of children this table has
			while (bb < numConnectedTables)
			{
				
				if (!(newRelations[tableCount][bb]==null))
				{
					//Table tableCount has a one to many relation with b
					numelts++;
					numParents[bb]++;
					//	System.out.println("not null");
				
				}
				
				//System.out.println(numelts);
				//System.out.println(bb+ " " +tableCount +" "+ numParents[bb]);
				bb++;
			}//end inner while (bb)

			children[tableCount]= new int[numelts];
			tableCount++;
		}
		 //end WHILE (tableCount)
	
		

		 //FINDING OUT THE ROOT NODE
		
		int maxParents=0; // the maximum number of parents any table has
		int croot = 0; //current root node
		for (int i=0; i < numConnectedTables; i++ )
			{
				parents[i]=new int[numParents[i]];
				//System.out.println(i +" has " + numParents[i] + " parents");
				if (numParents[i]>maxParents)
				{
				maxParents=numParents[i];
				croot=i;
				}
			}
		root=croot;
		//System.out.println("Root node=" +croot);
		tableCount=0;	
	
		while (tableCount < numConnectedTables)
		{	
			int bb=0; //our counter for columns
			int numelts=0; //count the number of children this table has
				
			while (bb<numConnectedTables)
			{
				//System.out.println(bb+ " " +tableCount +" "+ numParents[bb]);
			
				if (!(newRelations[tableCount][bb]==null))
				{
					//Table tableCount has a one to many relation with b
					children[tableCount][numelts]=bb;
					numParents[bb]--;
					parents[bb][numParents[bb]]=tableCount;
					numelts++;
				}//end if
				bb++;
			}//end while(bb)
			tableCount++;
		}//end while(tableCount)

		// Form relations and put in hashtable
		makeRelations();

		//Parents and Children are now DONE
		schemaCreated = true;
	}// try
	catch (SQLException e)
	{
		//e.printStackTrace();
		System.err.println(e.toString());
		String errorMessage = 
			"Unable to retrieve the table relationships from the " +
			"database.\nThis may be an MS Access database.\n\n" +
			"Snap will not be able to convert keys for events fired " +
			"between related tables.";
		//System.err.println(errorMessage);
		MsgBox msg = new MsgBox("DatabaseSchema Error", errorMessage, false);
		schemaCreated = false;
	}
} //end constructor DatabaseSchema


private void setTables()
{
	try
	{
		rs = dmd.getTables(catalog, null, null, tableTypes);

		// We store the list of tables in this Vector
		while(rs.next())
			tables.add(rs.getString("TABLE_NAME"));
	}
	catch(SQLException e)
	{
	}
}

private void setColumns()
{
	Enumeration e = tables.elements();
	while (e.hasMoreElements())
	{
		String tableName = (String) e.nextElement();
		columns.put(tableName, retrieveColumns(tableName));
	}
}

private Vector retrieveColumns(String table)
{
	Vector columns = new Vector();
	try 
	{
		rs = dmd.getColumns(catalog, null, table, null);
		while (rs.next())
			columns.add(rs.getString("COLUMN_NAME"));
	}
	catch(Exception e)
	{
		// System.err.println("Unable to call getColumns()");
	}
	return columns;
}

private void setPrimaryKeys()
{
	Enumeration e = tables.elements();
	while (e.hasMoreElements())
	{
		String tableName = (String) e.nextElement();
		primaryKeys.put(tableName, retrievePrimaryKeys(tableName));
	}
}

private Vector retrievePrimaryKeys(String table)
{
	Vector keys = new Vector();

	try
	{
		rs = dmd.getPrimaryKeys(catalog, null, table);
		while (rs.next())
			keys.add(rs.getString("COLUMN_NAME"));
	}
	catch(SQLException e)
	{	// Unable to getPrimaryKeys, try to find the first column
		keys.add(getFirstColumn(table));
	}

	return keys;
}

private String getFirstColumn(String table)
{
	Vector v = getColumns(table);
	if (v == null || v.isEmpty())
		return null;
	else
		return (String) v.elementAt(0);
}

/**
 * Gets the primary key for a particular table.  Note: This only returns a
 * single String as the primary key.  However, the primary key may actually
 * be made up of a combination of columns.
 */
public String getPrimaryKey(String table)
{
	Vector keys = (Vector) primaryKeys.get(table);
	if (keys == null || keys.isEmpty())
		return null;
	else
	{	// Returns the first primary key
		return (String) keys.elementAt(0);
	}
}

/** Returns a Vector of table names */
public Vector getTables()
{
	return tables;
}

/** Returns a Vector of column names for a given table */
public Vector getColumns(String table)
{
	return (Vector) columns.get(table);
}

/**
 * Makes the relations among all tables and
 * puts them in a hashtable.
 */
private void makeRelations()
{
	// Traverse the tables for names
	for (int i=0;i < newTables.length; i++ )
	{
		for (int j=0; j < children[i].length; j++)
		{
			String OneManyRelation = newTables[i].trim() + 
				newTables[children[i][j]].trim();
			// Caching the table relationships
			oneToManyTableRelationship.put(OneManyRelation, new Object());
		}
		for (int j=0; j < newTables.length; j++)
		{
			String primaryKeyI = getPrimaryKey(newTables[i]);
			String primaryKeyJ = getPrimaryKey(newTables[j]);
			if(primaryKeyI != null && primaryKeyJ != null &&
				primaryKeyI.equals(primaryKeyJ))
			{
				String oneToOneRelationship = newTables[i] + newTables[j];
				// Caching the table relationships
				oneToOneTableRelationship.put(
					oneToOneRelationship, new Object());
			}
		}
	}
}

/**
 * Check from hashtable and return the relation.
 */
public int getRelation(String tableSrc, String tableDest)
{
	String src = tableSrc.trim();
	String dest = tableDest.trim();

	if (!schemaCreated)
		return NO_RELATION;
	else if (oneToOneTableRelationship.containsKey(src + dest))
		return ONE_TO_ONE;
	else if (oneToManyTableRelationship.containsKey(src + dest))
		return ONE_TO_MANY;		
	else if (oneToManyTableRelationship.containsKey(dest + src))
		return MANY_TO_ONE;
	else
		return NO_RELATION;
}


}