
package snap.graph;

import snap.database.*;  // for the Query object
import snap.adapter.*; // for the TechnologyAdapter
import snap.*;

import java.util.*;

/** A node in the Graph; the Node contains all of the information for a
  * Visualization.  The Node is created by the UI.  It contains details about
  * the visualization, a point to the adapter, and information about the  
  * connections.
  *
  * @author Nathan Conklin - nathan@conklinfamily.net
  * @author Chris North - north@cs.vt.edu
  * @author Varun Saini - vsaini@vt.edu
  * @author Umer Farooq - ufarooq@vt.edu
  * @author Aniket Prabhune - aprabhun@vt.edu
  */
public class Node
{

private Graph graph;
private String name;
private Query query;
private TechnologyAdapter adapter;


/** The set of one direction connections stored as a hashtable of
NodeActionPairs.  The hashtable is keyed on by the action of the 
current node.  Its values are a Vector of one direction links stemming 
out from the current node.
	
	"select"	->	( (VisA, "select")
					  (VisB, "load")	)
	"mouseOver"	->	((VisA, "mouseOver"))
	
*/

private Hashtable connections = new Hashtable(); 

public Node(Graph graph, String name, Query query, TechnologyAdapter adapter)
{
	this.graph = graph;
    setName(name);
    this.query = query; // setQuery() not called becase it calls adapter.load()
    setTechnologyAdapter(adapter);
}

public void removeConnection(String action, NodeActionPair pair)
{
	Vector v = (Vector) connections.get(action);
	if (v != null)
	{
		for (int i = 0; i < v.size(); i++ )
		{
			NodeActionPair element = (NodeActionPair) v.elementAt(i);
			if (element.getAction().equals(pair.getAction()) &&
				element.getNode() == pair.getNode())
			{
				v.remove(i);
				return;
			}
		}
	}
}

public void addConnection(String action, NodeActionPair pair)
{
	Vector v = (Vector) connections.get(action);
	if (v == null)
	{
		v = new Vector();
		connections.put(action, v);
	}
	v.add(pair);
}

/** Returns an Vector of NodeActionPairs */
public Vector getConnections(String action)
{
	Vector v = (Vector) connections.get(action);
	if (v == null)
		return v = new Vector();
	return v;
}

public Enumeration getSupportedActions()
{
	return adapter.getSupportedActions();
}

public void setName(String name)
{
    this.name = name;
}

public String getName()
{
    return name;
}

public void setGraph(Graph graph)
{
    this.graph = graph;
}

public Graph getGraph()
{
    return graph;
}

public void setQuery(Query query)
{
	// Checks to see if it is a new query being set
	if(!this.query.equals(query))
	{
	    this.query = query;
		if (!query.isEmpty())  // Checks that the new Query is not blank
			adapter.load(query);
	}
}

public Query getQuery()
{
    return query;
}

public void setTechnologyAdapter(TechnologyAdapter adapter)
{
    this.adapter = adapter;
	adapter.setNode(this);
}

public TechnologyAdapter getTechnologyAdapter()
{
    return adapter;
}

}

