package snap.ui.diagram;

import snap.*;
import snap.coordination.*;
import snap.ui.*;

import snap.graph.*;
import snap.adapter.*;
import snap.database.*;

import javax.swing.*;
import java.sql.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.net.*;

/**
  * DiagramPanel is the component object used to display layout information 
  * such as visualization icons and their associated connections.  This 
  * object encapsulates all the necessary component objects for layout 
  * design including the dialog windows.  It extends JComponent and 
  * implements MouseListener, MouseMotionListener and ActionListener to 
  * allow for mouse controls and actions.   It also contains the
  * actions for adding, editing, and deleting layout components.
  *
  * @author Nathan Conklin - nathan@conklinfamily.net
  * @author Matt Clement - macleme2@vt.edu
  * @author Kiran Indukuri - kindukur@vt.edu
  * @author Chris North - north@cs.vt.edu
  */
public class DiagramPanel extends JComponent
implements MouseListener, MouseMotionListener,
ActionListener, GraphChangeEventListener
{
   
protected Snap snap;

// list of ALL Visualizations
private Vector visualizationList = new Vector();

// list of ALL Links
private Vector linkList = new Vector();

// Used for the click and drag events
private Visualization currentVisualization = null;

// Keeps track of when a link is being created.  Is true when the link
// has been started but the second component has not been chosen.
private boolean addingNewLink = false;

private Visualization newLinkFirstVisualization = null;
private Visualization newLinkSecondVisualization = null;
private Link newLink = null;

//Holds link to modify from list of multiple links
private Link linkToModify = null;
private Coordination coordinationToModify = null;

private JPopupMenu singleVisualizationPopup;
private JPopupMenu multipleVisualizationPopup;

// Variables used in popups for editing and removing coordinations
private JPopupMenu linkPopup;
private JMenu editCoordinationPopup;
private JMenu deleteCoordinationPopup;

// Stores click point from Right clicks for ActionEvents
private Point popupMenuStartPoint;

// Difference between point-of-click and the icon position
// Used so that the icon can be placed relative to the mouse location
private int xDiff, yDiff;

private CoordinationSettings coordinationSettingsWindow;
private VisualizationSettings visualizationSettingsWindow;

protected Point newVisualizationPoint = new Point(60, 60);
protected Dimension	preferredSize = new Dimension(300,300);

private final static int EDIT_COORDINATION = 0;
private final static int REMOVE_COORDINATION = 1;

/**
 * Class constructor
 * 
 * @param snap		    the snap parent object
 * @param dataConnect   the connection to the database data (user data)
 */
public DiagramPanel(Snap snap)
{          
	this.snap = snap;
	xDiff = yDiff = 0;
	jbInit();               //Initialize JComponent
}

/**
 * Initializes the DiagramPanel JComponent with the appropriate values.  
 * This includes adding the appropriate listenenrs and setting up the 
 * context menus for the object.
 */
protected void jbInit()
{
	addMouseListener(this);                 //Add Listeners
	addMouseMotionListener(this);
	this.setPreferredSize(preferredSize);   //Set Preferred Size

	// Create Popup Menus:
	// Visualization Popup menu - show on right click of visualization
	// If only one visualization is available, no links can be made
	JMenuItem singleEditMenuItem = new JMenuItem("View Properties");
	singleEditMenuItem.addActionListener(this);
	singleVisualizationPopup = new JPopupMenu();
	singleVisualizationPopup.setName("One Visualization Menu");
	singleVisualizationPopup.add(singleEditMenuItem);
	
	JMenuItem multipleEditMenuItem = new JMenuItem("View Properties");
	multipleEditMenuItem.addActionListener(this);
	JMenuItem multipleAddMenuItem = new JMenuItem("Add Coordination");
	multipleAddMenuItem.addActionListener(this);
	multipleVisualizationPopup = new JPopupMenu();
	multipleVisualizationPopup.setName("Multiple Visualization Menu");
	multipleVisualizationPopup.add(multipleEditMenuItem);
	multipleVisualizationPopup.add(multipleAddMenuItem);

	repaint();
}

/** Finds a new point to place the visualization. This method is
  * only public within the package. */
Point getNewVisualizationPoint()
{
	int increment = 60;
	int xMax = 200;
	int yMax = 300;
	Point newPoint = new Point(
		newVisualizationPoint.x, newVisualizationPoint.y);
	newVisualizationPoint.x = newVisualizationPoint.x + increment;
	if(newVisualizationPoint.x > xMax)
	{
		newVisualizationPoint.x = increment;
		newVisualizationPoint.y = newVisualizationPoint.y + increment;
		if(newVisualizationPoint.y > yMax)
		{
			newVisualizationPoint.y = increment;
		}
	}
	return newPoint;
}

/**
 * Finds the Visualization object being painted that is located at the
 * point indicated by the passed parameter.  If no such visualization is
 * found, null is returned.
 *
 * @param p             the point in question
 *
 * @returns returns the Visualization whose boundary contains the point
 *          in question
 */
private Visualization findVisualization(Point p)
{
	Visualization visualization;
	Enumeration e = visualizationList.elements();
	while(e.hasMoreElements())
	{
		visualization = (Visualization) e.nextElement();
		if(visualization.containsPoint(p))
			return visualization;
	}
	return null;
}

/**
 * Finds the Link object that is located at the point indicated 
 * by the passed parameter.  If no such connection is found, null is
 * returned.
 */    
private Link findLink(Point p)
{
	Link link;
	Enumeration e = linkList.elements();
	while(e.hasMoreElements())
	{
		link = (Link) e.nextElement();
		if(link.containsPoint(p))
			return link;
	}
	return null;
}

/** Finds the Link object that connects the two Visualizations. */
private Link findLink(Visualization testStart, Visualization testEnd)
{
	Enumeration e = linkList.elements();
	while(e.hasMoreElements())
	{
		Link link = (Link) e.nextElement();
		if (	(link.getStartVisualization().equals(testStart) &&
				link.getEndVisualization().equals(testEnd))
			||
				(link.getStartVisualization().equals(testEnd) &&
				link.getEndVisualization().equals(testStart))
			)
		{
			return link;
		}
	}
	return null;
}

/**
 * Handles the repainting of all the visualizations and connections 
 * currently being displayed in the layout.
 *
 * @param g             the graphics context to use for painting
 */
public void paintComponent(Graphics g)
{
	Visualization loopVisualization;
	Link    loopLink;

	Enumeration e = linkList.elements();
	while(e.hasMoreElements())
	{
		loopLink = (Link)e.nextElement();
		loopLink.paintComponent(g);
	}

	e = visualizationList.elements();
	while(e.hasMoreElements())
	{
		loopVisualization = (Visualization) e.nextElement();
		loopVisualization.paintComponent(g);
	}

	if(addingNewLink)
	{
		newLink.paintComponent(g);
		newLink.getStartVisualization().paintComponent(g);
		newLink.getEndVisualization().paintComponent(g);
	}
}

/**
 * Interprets mouse clicked events and determines the appropriate UI 
 * related actions to perform.  If a link is in the process of being 
 * created, this method helps regulate that process and is responsible for 
 * initiating the connection completion sequence when the process is
 * complete.
 *
 * @param mouseEvent the mouse event to be handled
 */
public void mouseClicked(MouseEvent mouseEvent)
{
	int x = mouseEvent.getX();
	int y = mouseEvent.getY();

	if( addingNewLink && newLinkFirstVisualization != null)
	{
		newLinkSecondVisualization = findVisualization(mouseEvent.getPoint());
		if (newLinkSecondVisualization != null)
			endAddCoordination();
		else
		{
			addingNewLink = false;
			newLinkFirstVisualization = null;
			newLink = null;
			repaint();
		}
	}
}


/**
 * Handles the continuous drawing and animation of a temporary connection
 * as a result of mouse dragged events.  It also ensures that the 
 * visualizations are drawn on top of the connections to present the
 * correct visual appearance.
 *
 * @param mouseEvent the mouse event to be handled
 */
public void mouseDragged(MouseEvent mouseEvent)
{
	if(currentVisualization != null)
	{
		currentVisualization.erase(this.getGraphics());
		if(currentVisualization.isLinked())
		{
			Enumeration e = currentVisualization.getLinks().elements();
			while (e.hasMoreElements())
			{
				Link link = (Link) e.nextElement();
				link.erase(this.getGraphics());
				if(currentVisualization == link.getStartVisualization())
					link.setStartLocation(
						new Point(currentVisualization.getCenterPoint()));
				else
					link.setEndLocation(
						new Point(currentVisualization.getCenterPoint()));
				//link.paintComponent(this.getGraphics());
			}
		}
		
		currentVisualization.setLocation( new Point(
			mouseEvent.getX() - xDiff, mouseEvent.getY() - yDiff));
		//currentVisualization.paintComponent(this.getGraphics());
		adjustPanelSize(mouseEvent.getPoint());
	}
	repaint();
}

/**
 * Handles all mouse moved events, but particularly is involved with the
 * continuous drawing and animation of a new Link as a result of these
 * mouse moved events.
 *
 * @param mouseEvent the mouse event to be handled
 */    
public void mouseMoved(MouseEvent mouseEvent)
{
	if(addingNewLink)
	{
		newLink.erase(this.getGraphics());
		newLink.setEndLocation(
		  new Point(mouseEvent.getPoint()));
		repaint();
	}
}

/**
 * Initiates the sequence of events to manage the popup menus as a result
 * of mouse release events.
 *
 * @param mouseEvent the mouse event to be handled
 */    
public void mouseReleased(MouseEvent mouseEvent)
{
	if(mouseEvent.isPopupTrigger())
		showPopupMenu(mouseEvent);
}

/**
 * Initiates the sequence of events to manage the popup menus as a result
 * of mouse pressed events and establishes a current visualization
 * otherwise.
 *
 * @param mouseEvent the mouse event to be handled
 */    
public void mousePressed(MouseEvent mouseEvent) {
	if(mouseEvent.isPopupTrigger())
		showPopupMenu(mouseEvent);
	else
	{
		currentVisualization = findVisualization(mouseEvent.getPoint());
		if(currentVisualization != null)
		{
			// A visualization is selected, the click offset must be determined
			xDiff = mouseEvent.getX() - 
			  (int)(currentVisualization.getLocation().getX());
			yDiff = mouseEvent.getY() - 
			  (int)(currentVisualization.getLocation().getY());
		}
	}
}

/**
 * Handles the mouseEntered event.  Retained as part of implementations.
 *
 * @param mouseEvent    the mouse event thrown
 */
public void mouseEntered(MouseEvent mouseEvent) {}
    
/**
 * Handles the mouseExited event.  Retained as part of implementations.
 *
 * @param mouseEvent    the mouse event thrown
 */    
public void mouseExited(MouseEvent mouseEvent) {}
    
/**
 * Adjusts the preferred size of the panel in case the user drags a
 * Visualization object outside the current bounds of the screen.
 *
 * @param newPoint      the current location of the object to include
 */
private void adjustPanelSize(Point newPoint)
{
	Dimension tempDimension = this.getSize();
	if(newPoint.getX() >= tempDimension.getWidth())
		tempDimension.setSize(
			newPoint.getX() + 50, tempDimension.getHeight());
	if(newPoint.getY() >= tempDimension.getHeight())
		tempDimension.setSize(
			tempDimension.getWidth(), newPoint.getY() + 50);
	this.setPreferredSize(tempDimension);
}

/**
 * Handles all actions and events related to the drawing of a layout sent
 * by way of mouse events or through other method calls.  These atomic
 * events make up the basic operations that can be performed.
 * Iterpretation of the events to determine not only the type of event that 
 * was sent, but where the event originated is pivotal to the operation of
 * this method.
 *
 * @param actionEvent   the event to process
 */
public void actionPerformed(ActionEvent actionEvent)
{
	if(addingNewLink)   //Don't even bother
		return;

	String sourceText = actionEvent.getActionCommand(); //Get source text

	if(sourceText.equals("View Properties"))
		editVisualization(findVisualization(popupMenuStartPoint));
	else if(sourceText.equals("Add Coordination"))
		startAddCoordination();
	else
		try
		{
			Component source = (Component) actionEvent.getSource();
			if(editCoordinationPopup.isMenuComponent(source))
				updateCoordination(sourceText, EDIT_COORDINATION);
			else if(deleteCoordinationPopup.isMenuComponent(source))
				updateCoordination(sourceText, REMOVE_COORDINATION);
		}
		catch(Exception e)
		{
			// Unable to cast the getSource() object into a Component
		}

}


/** Called by the edit and remove popup menus to edit or remove a
  * Coordination */
private void updateCoordination(String name, int editOrRemove)
{
	Link updateLink = null;
	Coordination updateCoordination = null;

	boolean found = false;

	Enumeration e = linkList.elements();
	while(!found && e.hasMoreElements())
	{
		Link link = (Link) e.nextElement();
		Enumeration coordEnum = link.getCoordinations().elements();
		while (!found && coordEnum.hasMoreElements())
		{
			Coordination coordination = (Coordination) coordEnum.nextElement();
			if (coordination.getName().equals(name))
			{
				updateLink = link;
				updateCoordination = coordination;
				found = true;
			}
		}
	}
	if (found  && editOrRemove == EDIT_COORDINATION)
	{
		coordinationSettingsWindow = new CoordinationSettings(
			updateLink, updateCoordination, this,
			snap.getDatabaseManager().getDatabaseSchema());
	}
	else if (found  && editOrRemove == REMOVE_COORDINATION)
	{
		removeCoordination(updateLink, updateCoordination);
		if (updateLink.getCoordinations().size() == 0)
			removeLink(updateLink);
	}
}

/**
 * Ends the process of adding a new connection.
 */
private void endAddCoordination()
{
	newLink.setEndVisualization(newLinkSecondVisualization);
	
	Link alreadyConnectedLink = findLink(
		newLinkFirstVisualization, newLinkSecondVisualization);
	if(alreadyConnectedLink ==  null)
	{	// Add a new link between the two components
		linkList.add(newLink);
		newLinkFirstVisualization.addLink(newLink);
		newLinkSecondVisualization.addLink(newLink);
	}
	else
		newLink = alreadyConnectedLink;

	coordinationSettingsWindow =
		new CoordinationSettings(newLink, this,
			snap.getDatabaseManager().getDatabaseSchema());

	newLink = null;
	addingNewLink = false;
	repaint();
}

/**
 * Begins the process of adding a new connection.
 */
private void startAddCoordination()
{
	if(popupMenuStartPoint != null)
		newLinkFirstVisualization = findVisualization(popupMenuStartPoint);
	if(newLinkFirstVisualization != null)
	{
		addingNewLink = true;
		newLink = new Link(
			newLinkFirstVisualization.getCenterPoint(),
			newLinkFirstVisualization, 
			newLinkFirstVisualization.getCenterPoint(), 
			newLinkFirstVisualization, this);
	}
	else
		snap.getFeedbackPanel().setText(
		"startAddCoordination() called without a starting visualization.");
}
    
/**
 * Edits an existing visualization.  This represents one of the atomic 
 * layout actions.
 */    
private void editVisualization(Visualization visualization)
{
	visualizationSettingsWindow = new VisualizationSettings(
		visualization, snap.getDatabaseManager().getDatabaseSchema());
}

/**
 * Displays the appropriate popup menu determined by the context of the 
 * item that was selected or clicked on.  Each popup menu has commands and
 * options available for that specific item.
 *
 * @param e                 the mouse event to be processed
 */
private void showPopupMenu(MouseEvent e)
{
	if(!addingNewLink) //Disallows popup menus while add a link
	{
		popupMenuStartPoint = new Point(e.getPoint());
		
		if(findVisualization(e.getPoint()) != null) // Popup on Visualization
		{
			// If only one visualization is available, no links can be made
			if(visualizationList.size() == 1)
				singleVisualizationPopup.show(
					e.getComponent(), e.getX(), e.getY());
			else
				multipleVisualizationPopup.show(
					e.getComponent(), e.getX(), e.getY());
		}
		else if(findLink(e.getPoint()) != null) // Popup on a Link
		{
			// Connection Popup menu - show on right click of connection
			linkPopup = new JPopupMenu();
			linkPopup.setName("ConnectionMenu");
			editCoordinationPopup = new JMenu("Edit Coordination");
			deleteCoordinationPopup = new JMenu("Delete Coordination");	

			editCoordinationPopup.addActionListener(this);
			deleteCoordinationPopup.addActionListener(this);

			linkPopup.add(editCoordinationPopup);
			linkPopup.add(deleteCoordinationPopup);

	
			Link linkRightClicked = findLink(e.getPoint());
			//Add connections to delete connections popup list....

			deleteCoordinationPopup.removeAll();
			
			Enumeration enum = linkRightClicked.getCoordinations().elements();
			while(enum.hasMoreElements())
			{
				Coordination c = (Coordination) enum.nextElement();
				JMenuItem item = new JMenuItem(c.getName());
				item.addActionListener(this);
				deleteCoordinationPopup.add(item);
			}

			
			editCoordinationPopup.removeAll();
			enum = linkRightClicked.getCoordinations().elements();
			while(enum.hasMoreElements())
			{
				Coordination c = (Coordination) enum.nextElement();
				JMenuItem edititem = new JMenuItem(c.getName());
				edititem.addActionListener(this);
				editCoordinationPopup.add(edititem);
			}

			//updateDeletePopup(linkRightClicked);
			linkPopup.show(e.getComponent(), e.getX(), e.getY());
		}
	}
}


public void graphChanged(GraphChangeEvent gce)
{
	if(gce.getEventType() == GraphChangeEvent.NODE_ADDED)
	{
		// "Add Visualization"
		Visualization vis = new Visualization((UINode)gce.getNode(), this);
		visualizationList.add(vis);
		editVisualization(vis);
	}
	else if(gce.getEventType() == GraphChangeEvent.NODE_REMOVED)
	{
		Visualization toRemove = null;
		Enumeration e = visualizationList.elements();
		while (e.hasMoreElements())
		{	// Find the visualization to remove
			Visualization vis = (Visualization) e.nextElement();
			if (vis.getUINode() == gce.getNode())
				toRemove = vis;
		}

		if (toRemove != null)
		{	// Remove the visualization
			visualizationList.remove(toRemove);

			// Copy the links to remove because we can't actually
			// remove the links while iterating across the Vector
			e = toRemove.getLinks().elements();
			Vector links = new Vector(); // Holds a copy of the links
			while(e.hasMoreElements())
				links.add(e.nextElement());

			// Remove all of the links
			e = links.elements();
			while (e.hasMoreElements())
			{
				Link link = (Link) e.nextElement();
				removeLink(link);
			}
		}
	}
	repaint();
}

/** Removes the link from the UI and from the CoordinationGraph */
public void removeLink(Link link)
{
	// Remove the link from the UIGraph
	link.getStartVisualization().removeLink(link);
	link.getEndVisualization().removeLink(link);
	linkList.remove(link);

	// Copy the coordinations to remove because we can't actually
	// remove the coordinations while iterating across the Vector
	Enumeration e = link.getCoordinations().elements();
	Vector coordinations = new Vector(); // Holds a copy of the links
	while(e.hasMoreElements())
		coordinations.add(e.nextElement());

	e = coordinations.elements();
	while (e.hasMoreElements())
		removeCoordination(link, (Coordination) e.nextElement());
	repaint();
}

/** Removes the coordination from the UI and from the CoordinationGraph */
private void removeCoordination(Link link, Coordination coordination)
{
	// Removes the coordination from the UIGrph
	link.getCoordinations().remove(coordination);
	
	// Gets information about the graph Nodes
	UINode nodeA = (UINode) link.getStartVisualization().getUINode();
	String actionA = coordination.getStartAction();
	UINode nodeB = (UINode) link.getEndVisualization().getUINode();
	String actionB = coordination.getEndAction();

	try
	{
		// Removes the coordination from the CoordinationGraph's data structure
		nodeA.getGraph().getMonitor().acquireWrite();
		// Begin critical section
		// --------------------------------------------------------
			nodeA.removeConnection(actionA,	new NodeActionPair(nodeB, actionB));
			nodeB.removeConnection(actionB,	new NodeActionPair(nodeA, actionA));
		// --------------------------------------------------------
		// End critical section
		nodeA.getGraph().getMonitor().releaseWrite();
	}
	catch(InterruptedException e)
	{
		System.err.println(
			"Unable to obtain lock for the graph: " + e.toString());
	}
}


};
