package snap.ui.diagram;

import java.awt.*;
import java.util.Vector;

/**
  * Link is the UI abstraction of a Link between two visualizations.
  * The Link object encapsulates all the information needed to build
  * the UI representation of a connection within a layout.
  *
  * @author Nathan Conklin - nathan@conklinfamily.net
  * @author Matt Clement - macleme2@vt.edu
  * @author Chris North - north@cs.vt.edu
  */
public class Link
{

private DiagramPanel        container;	//containing object
private Point               startPoint,	//start pt. of conneciton
							endPoint;	//end pt. of connection
private Visualization		startViz,	//starting visualization
							endViz;		//ending visualization
private Polygon             clickArea;
private Vector				coordinations; // contains Coordinations

/**
 * Class constructor
 * 
 * @param start         the starting point of the UI representation
 * @param visStart      the visualization where the connection starts
 * @param end           the ending point of the UI representation
 * @param visEnd        the visualization where the connection ends
 * @param thisContainer the containing object, in this case DiagramPanel
 *
 */
public Link(Point start, Visualization startViz,
	Point end, Visualization endViz, DiagramPanel container)
{
	startPoint     = start;
	endPoint       = end;
	this.startViz  = startViz;
	this.endViz    = endViz;
	this.container = container;        
	clickArea      = null;
	coordinations  = new Vector();
}

/**
 * Retrieves the name of the Link object.
 *
 * @return returns the name of the Link object as type String
 */
public String getName()
{
	return startViz.getUINode().getName() + " - " +
		endViz.getUINode().getName();
}

/**
 * Returns a Vector of Coordination objects.  Each coordination is
 * the link between one action and other.
 */
public Vector getCoordinations()
{
	return coordinations;
}

/**
 * Retrieves the location where the Link object starts
 * 
 * @return returns the location as type Point where the object starts
 */    
public Point getStartLocation()
{
	 return startPoint;
}    

/**
 * Retrieves the location where the Link object ends.
 * 
 * @return returns the location as type Point where the object ends
 */    
public Point getEndLocation()
{
	 return endPoint;
}

/**
 * Retrieves the Visualization object at the beginning of the
 * connection
 *
 * @return retuns the Visualization object at the beginning of the
 * connection
 */
public Visualization getStartVisualization()
{
	return startViz;
}

/**
 * Retrieves the Visualization object at the end of the connection.
 *
 * @return retuns the Visualization object at the end of the
 * connection
 */    
public Visualization getEndVisualization()
{
	 return endViz;
}

/**
 * Sets the visualization where the connection should begin to that
 * specified by the parameter.
 *
 * @param newStart      the new starting visualization
 */
public void setStartVisualization(Visualization newStart)
{
	startViz = newStart;
}

/**
 * Sets the visualization where the connection should end to that
 * specified by the parameter.
 *
 * @param newStart      the new ending visualization
 */    
public void setEndVisualization(Visualization newEnd)
{
	endViz = newEnd;
}   

/**
 * Sets the location where the Link should begin painting to a
 * point passed in as a parameter.
 *
 * @param newPoint      the new beginning location for the Link
 */
public void setStartLocation(Point newPoint)
{
	startPoint.setLocation(newPoint);
	adjustClickableArea();      //clickable area must be adjusted afterward
}

/**
 * Sets the location where the Link should stop painting to a
 * point passed in as a parameter.
 *
 * @param newPoint      the new ending location for the Link
 */    
public void setEndLocation(Point newPoint)
{
	endPoint.setLocation(newPoint);
	adjustClickableArea();      //clickable area must be adjusted afterward
}

/**
 * Adjusts the area around the Link object, the "clickable area,"
 * to the appropriate values after adjusting the starting or ending
 * locations.
 */
private void adjustClickableArea()
{
	int polygonPoints          = 4; //number of points in the area shape
	int clickableThickness     = 5; //thickness of the clickable area
	int yPoints[] = {(int)(startPoint.getY() - clickableThickness), 
					(int)(endPoint.getY() - clickableThickness), 
					(int)(endPoint.getY() + clickableThickness), 
					(int)(startPoint.getY() + clickableThickness)};
	int xPoints[] = {(int)(startPoint.getX() - clickableThickness), 
					(int)(endPoint.getX() - clickableThickness), 
					(int)(endPoint.getX() + clickableThickness), 
					(int)(startPoint.getX() + clickableThickness)};
	clickArea    = new Polygon(xPoints, yPoints, polygonPoints);        
}

/**
 * Calculates and returns the length of the Link object based on
 * simple geometric principles.
 *
 * @return returns the length of the UI representation of the
 * Link object
 */
private int getLength()
{
	return (int)(Math.sqrt(
				Math.abs(startPoint.getX()-endPoint.getX()) * 
				Math.abs(startPoint.getX()-endPoint.getX()) + 
				Math.abs(startPoint.getY()-endPoint.getY()) * 
				Math.abs(startPoint.getY()-endPoint.getY())
		));
}

/**
 * Determines whether a point given as a parameter is within the clickable
 * region of the Link object and returns boolean values as
 * appropriate.
 *
 * @param testPoint     the point to test for containment
 *
 * @return returns true if the point in question is within the clickable
 *         area of the Link and false otherwise.
 */
public boolean containsPoint(Point testPoint)
{
	if(clickArea.contains(testPoint))
		return true;
	else
		return false;        
}

/**
 * Draws the iconic represntation of the Link object at its
 * location.
 * 
 * @param g             the graphics context to use for painting
 */    
public void paintComponent(Graphics g)
{
	g.setColor(Color.black);
	g.drawLine(startPoint.x, startPoint.y, endPoint.x, endPoint.y);
}

/**
 * Erases the iconic representation of the Link object.  This
 * is accomplished by a simple overlay of appropriate background colors.
 *
 * @param g             the graphics context to use for painting
 */    
public void erase(Graphics g)
{
	g.setColor(container.getBackground());
	g.drawLine(startPoint.x, startPoint.y, endPoint.x, endPoint.y);        
}

}
