/*  XWAND - Written by Andrew A. Ray (anray2@vt.edu) [August 2002]
 *  	With help from 
 *  	               John Kelso        (Ideas and data smoothing)
 *  		       Ron Kriz          (Support and professional opinion!)
 *		       Lance Arsenault   (Lecturing)
 *                     Mark J. Soltys II (Message passing)
 *                     Robert Stein  (X server control interface)
 *
 *     This program basically allows you to use a wand / other
 *     method of control to move your mouse around and simulate
 *     mouse clicks.  
 *
 *     When XWand starts it reads information from a configuration 
 *     file about what gui to start, where files are located, what 
 *     trigger button to use, etc...
 *
 *     XWand sits and listens for a trigger button, and then will start
 *     a user defined gui.  The pid of the gui is stored in the parent
 *     so the parent can kill the child at any time
 *
 *     When the gui is started xwand simulates mouse clicks / releases
 *     as defined in the configuration file
 *
 *     This is by no means a perfect example of how to code, but it serves
 *     its purpose and is rather flexible due to the config file. If you
 *     would like to improve on the code, feel free to.  If you have
 *     questions feel free to email me.  I may not have time to 
 *     help you, but if I do I will.
 */ 
#include <dtk.h>           //For message passing / getting of input
#include <string>          //For message passing
#include "FakeMouse.h"     //For the controlling the xserver
#include "Messenger.h"     //For sending messages
#include "xwandparser.h"   //For parsing the configuration file
#include "XWandCommands.h" //For the commands that xwand and its dso understand

//Global variables , for ease of use only
//If you want to be a code purist, change them
//to be passed into the functions
bool GuiStart;           //Is the gui started?
char* GuiPid;            //What is the gui's pid
FakeMouse*  Mouse;       //The xserver stuff
char** Environ;          //The environmental variables
XData TheData;           //The data from the  parser
dtkInButton* PollButton; //Used for polling the buttons
Messenger* Message;      //Used to send messages to the xwand dso
//End global variables

//Function prototypes
//For handling button presses
static void HandleButtons(dtkInButton* Button, dtkRecord_event* Event, const int& TriggerButton);
//For handling joystick movement
static void HandleJoyStick(dtkInValuator* JoyStick); 
//Run xwand
int main(int argc, char** argv, char** env)
{
	//The filename to open
	char* FileName = NULL;
	//Output debug information or not
	bool Verbose = false;
	//If we have arguments, handle them
	if (argc > 1)
	{
		for (int i=1; i<argc; i++)
		{	
			if ( !strcmp(argv[i], "-v"))
				Verbose = true;
			else
				FileName = argv[i];
		}
	}
	//Set the defaults for data
	TheData.Path = NULL;
	TheData.NumButtons = 0;
	TheData.Buttons = NULL;
	TheData.Display = "0.0";
	TheData.NumArgs = 0;
	TheData.App = NULL;
	TheData.NumActions = 0;
	TheData.Actions = NULL;
	TheData.InDir = false;
	TheData.JoySeg = false;
	TheData.ButtonSeg = false;
	TheData = XParse(FileName, Verbose);
	//Create a  messenger to send data to the
	//xwand dso
	Message = new Messenger("XWAND-DSO");
	
	//Used to get stuff out of shared memory
	dtkManager Manager;
	//Set the path for all of the dso stuff
	Manager.path(TheData.Path);
	//Get the cave input devices into the manager
	Manager.load("caveDTKInput");
	//Get the button state from the manager
	dtkInButton *Button = (dtkInButton *)Manager.get(TheData.ButtonSeg, DTKINBUTTON_TYPE);
	PollButton = (dtkInButton *)Manager.get(TheData.ButtonSeg, DTKINBUTTON_TYPE);

	//Queue the buttons
	Button->queue();
	//Get the joystick information
	dtkInValuator *JoyStick = (dtkInValuator *)Manager.get(TheData.JoySeg, DTKINVALUATOR_TYPE);
	//Get a record out of the manager so I can keep track of events (used for making sure I only get one button press)
	dtkRecord* Record = Manager.record();
	//An event used in conjunction with the record to make sure I only get one button press
	dtkRecord_event *event;
	//The trigger button used to start the application
	//I used the first button in the list as the master button
	//for ease of use
	int TriggerButton = -1;
	//Set the default
	if (TheData.Buttons !=NULL)
		TriggerButton = TheData.Buttons[0];	

	//Paranoia
	if (TriggerButton == -1)
	{
		printf("could not read trigger button, exiting\n");
		return -1;
	}
	//Information about the gui
	GuiStart = false;
	GuiPid = NULL;
	//Set the environmental variables
	Environ = env;
	//Get the xwand interface
	Mouse = new FakeMouse(TheData.Display);
	//Used to make sure we only get something when the button changes

	//While the state is ok and we are go for running
	while(Manager.state & DTK_ISRUNNING)
	{
		//Run the manager to get input
		if (Manager.frame()) return 1;
		//Loop to make sure I only get the right presses to the 
		//buttons
		while((event = Record->getNextEvent(Button)))
		{
			//Click the first button if a gui is started
			for (int i=0; i< TheData.NumActions; i++)
			{
				if (GuiStart && Button->read(i, event))
				{
					if (!strcmp(TheData.Actions[i], "press"))
						Mouse->Press(1);
					if (!strcmp(TheData.Actions[i], "release"))
						Mouse->Release(1);
					if (!strcmp(TheData.Actions[i], "click"))
						Mouse->Click(1);
					
				}
			}
					
			//Handle the button clicking
			HandleButtons(Button, event, TriggerButton);
		}
		//If the gui is started move the mouse if needbe
		if (GuiStart)
	      HandleJoyStick(JoyStick);
		//To keep it from being a cpu hog
		usleep(100000); 
	}
}

//If I get a button press , this function takes care of starting the gui / stopping the gui
static void HandleButtons(dtkInButton* Button, dtkRecord_event* Event, const int& TriggerButton)
{
	//Do we start / stop the navigation
	if (Button->read(TriggerButton, Event))
	{

		if (TheData.NumButtons > 1)
		{
			bool Pressed = false;
			for (int i=1; i<TheData.NumButtons; i++)
			{
				if (PollButton->read(TheData.Buttons[i]))
					Pressed=true;
				else
					Pressed=false;
			}
			//Wrong sequence of buttons!
			if (!Pressed)
				return;
					
		}
		if (GuiStart)
		{
			GuiStart = false;
			Message->SendCommand(STARTNAV);
		}
		else
		{
			Message->SendCommand(STOPNAV);
			GuiStart = true;
		}

		if (GuiStart)
			Mouse->SetPosition(2000,500);
			
			
/*
		//Kill the gui
		if (GuiStart) 
		{
			Message->SendCommand(STARTNAV);
			int killerpid = fork();
			if (killerpid ==0)
			{
				char* KillProg[4];
				KillProg[0]="kill";
				KillProg[1]="-9";
				KillProg[2]=GuiPid;
				KillProg[3]=NULL;
				//execvp(KillProg[0], KillProg);
				return;
			}
			else
			{
				GuiStart=false;
				delete [] GuiPid;
				GuiPid = NULL;
				return;
			}
		}
		//Create the gui
		else
		{
			Message->SendCommand(STOPNAV);
			int Pipe[2];	
			pipe(Pipe);
			int createpid = fork();
			if (createpid ==0)
			{
				int CPid = getpid();

				//Construct a char** to exec
				//It is 1 bigger b/c of a NULL
				//Terminator
				char** StartProg= new char*[TheData.NumArgs+1];
				//The program to execute
				StartProg[0]=TheData.App[0];
				for (int i=1; i<TheData.NumArgs; i++)
					StartProg[i]=TheData.App[i];
				StartProg[TheData.NumArgs] = NULL;

				write(Pipe[1], &CPid, sizeof(int));
				if (TheData.InDir)
					execve(StartProg[0], StartProg, Environ);
				else
					execvp(StartProg[0], StartProg);
			}
			else
			{
				//Get the id out of the pipe and then write it global memory
				int TempPid;
				read(Pipe[0], &TempPid, sizeof(int));
				GuiPid = new char[40];
				sprintf(GuiPid, "%d", TempPid);
				GuiStart= true;
				usleep(10000);
				//Put the mouse in the right position on the screen when we start
				Mouse->SetPosition(2000,2000);
			}
		}
*/
	}
}

//If the gui is started then take input from the joystick and move the mouse based on this
static void HandleJoyStick(dtkInValuator* JoyStick) 
{
	//Where I am at currently (has to be gotten from Mouse)
	int* CurrPos = new int[2];
	//Where the joystick is going to be
	float JoyPos[2];

	//Get the current mouses position
	CurrPos = Mouse->getMousePosition();
	//Get where the joystick is pointing
	JoyStick->read(JoyPos);
	//Update the mouse position
	if (JoyPos[0] !=0 || JoyPos[1] !=0)
	  {
	    //The updating factor
	    int ox = (int)(JoyPos[0]*JoyPos[0]*JoyPos[0]*25.f) ;
	    int oy = -(int)(JoyPos[1]*JoyPos[1]*JoyPos[1]*25.f) ;
	    
	    //Basically figure out what needs to be added
	    CurrPos[0] += ox ;
	    CurrPos[1] += oy ;
	    
	    //Set the mouses position
	    Mouse->SetPosition(CurrPos[0], CurrPos[1]);
	  }
	//Cleanup
	delete [] CurrPos;
}
